"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const typedi_1 = require("typedi");
const services_1 = require("../services");
const utils_1 = require("../utils");
const humanizeDuration = require('humanize-duration');
const deepObjectDiff = require('deep-object-diff');
class ActionSnapshot {
    constructor(source, idOrAlias, metadata, wd, idx, parameters) {
        this.source = source;
        this.idOrAlias = idOrAlias;
        this.metadata = metadata;
        this.wd = wd;
        this.idx = idx;
        this.parameters = parameters;
        this.ignoreChildFailure = false;
        this.duration = 0;
        this.createdAt = new Date();
        this.steps = [];
        this.successful = false;
        this.childFailure = false;
    }
    /**
     * Get human readable duration
     * @return {string}
     */
    getHumanReadableDuration() {
        return humanizeDuration(this.duration);
    }
    /**
     * Register log message
     * @param {string} message
     * @param {boolean} [error] indicated that log message is error
     * @param {boolean} [silent] if provided LogService will not be invoked
     */
    log(message, error = false, silent = false) {
        let prefix = ` -> [${this.idx}] [${this.source}] [${(this.metadata && this.metadata.$title) ||
            this.idOrAlias}]`;
        if (error) {
            prefix = prefix.red;
        }
        else {
            prefix = prefix.green;
        }
        const logMessage = `${prefix} ${message}`;
        if (!silent) {
            if (error) {
                typedi_1.Container.get(services_1.LogService).error(logMessage);
            }
            else {
                typedi_1.Container.get(services_1.LogService).info(logMessage);
            }
        }
        this.registerStep('log', message);
    }
    /**
     * Set ID of the action handler
     * @param {string} id
     */
    setActionHandlerId(id) {
        this.registerStep('actionHandler', id);
    }
    /**
     * Record start of the processing
     */
    start() {
        this.registerStep('start');
    }
    /**
     * Record when validation passed
     */
    validated() {
        this.registerStep('validated');
    }
    /**
     * Record successful execution state
     */
    success() {
        this.registerStep('success');
        this.completedAt = new Date();
        this.duration = this.completedAt.getTime() - this.createdAt.getTime();
        this.successful = this.ignoreChildFailure || !this.childFailure;
    }
    /**
     * Record when skipped execution
     */
    skipped() {
        this.registerStep('skipped');
        this.successful = this.ignoreChildFailure || !this.childFailure;
        this.completedAt = new Date();
        this.duration = this.completedAt.getTime() - this.createdAt.getTime();
    }
    /**
     * Record when execution failed with error
     * @param error
     */
    failure(error) {
        this.registerStep('failure', {
            message: (error && error.message) || /* istanbul ignore next */ error.toString(),
            code: error && error.code,
        });
        this.completedAt = new Date();
        this.duration = this.completedAt.getTime() - this.createdAt.getTime();
        this.successful = false;
        if (error && error.code) {
            this.errorCode = error.code;
        }
    }
    /**
     * Register child ActionSnapshot
     * @param {ActionSnapshot} snapshot
     */
    registerChildActionSnapshot(snapshot) {
        this.registerStep('child', snapshot);
        if (!snapshot.successful) {
            this.childFailure = true;
        }
    }
    /**
     * Get all the recorded steps
     * @return {IActionStep[]}
     */
    getSteps() {
        return this.steps;
    }
    /**
     * Set initial context state before applying any actions
     * @param {IContext} context
     */
    setInitialContextState(context) { } // tslint:disable-line
    /**
     * Register step. Use this method only when other methods do not suite your needs.
     * @param {string} type
     * @param payload
     */
    registerStep(type, payload) { } // tslint:disable-line
    /**
     * Register context
     * @param context
     */
    setContext(context) { } // tslint:disable-line
    /**
     * Register options
     * @param options
     */
    setOptions(options) { } // tslint:disable-line
}
exports.ActionSnapshot = ActionSnapshot;
class EnabledActionSnapshot extends ActionSnapshot {
    constructor(source, idOrAlias, metadata, wd, idx, parameters) {
        super(source, idOrAlias, metadata, wd, idx, parameters);
    }
    /**
     * @inheritdoc
     */
    setInitialContextState(context) {
        this.previousContextState = JSON.parse(JSON.stringify(utils_1.ContextUtil.toBase(context)));
    }
    /**
     * Register context
     * @inheritdoc
     */
    setContext(context) {
        // only selected fields should be exposed
        const newState = JSON.parse(JSON.stringify(utils_1.ContextUtil.toBase(context)));
        // extract diff
        const diff = deepObjectDiff.detailedDiff(this.previousContextState, newState);
        this.previousContextState = newState;
        const changes = Object.keys(diff.added).length + Object.keys(diff.deleted).length + Object.keys(diff.updated).length;
        // store diff
        /* istanbul ignore else */
        if (changes > 0) {
            this.registerStep('context', diff);
        }
    }
    /**
     * @inheritdoc
     */
    setOptions(options) {
        this.registerStep('options', JSON.parse(JSON.stringify(options)));
    }
    /**
     * @inheritdoc
     */
    registerStep(type, payload) {
        this.steps.push({
            type: type,
            createdAt: new Date(),
            payload: payload,
        });
    }
}
exports.EnabledActionSnapshot = EnabledActionSnapshot;
//# sourceMappingURL=ActionSnapshot.js.map