"use strict";
var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var FBLService_1;
Object.defineProperty(exports, "__esModule", { value: true });
const typedi_1 = require("typedi");
const Joi = require("joi");
const index_1 = require("./index");
const ActionHandlersRegistry_1 = require("./ActionHandlersRegistry");
const semver = require("semver");
const TemplateUtilitiesRegistry_1 = require("./TemplateUtilitiesRegistry");
const path_1 = require("path");
const utils_1 = require("../utils");
const schemas_1 = require("../schemas");
const commandExists = require('command-exists');
const requireg = require('requireg');
const fblVersion = require(process.env.FBL_ENV === 'test' ? '../../package.json' : '../../../package.json')
    .version;
let FBLService = FBLService_1 = class FBLService {
    constructor() {
        this.plugins = {};
        this.processedPlugins = {};
        this.reporters = {};
        // if plugin dependencies are not satisfied instead of throwing error log statements will be printed only
        this.allowUnsafePlugins = false;
        // if flow requirements are not satisfied instead of throwing error log statements will be printed only
        this.allowUnsafeFlows = false;
    }
    get flowService() {
        return typedi_1.default.get(index_1.FlowService);
    }
    get actionHandlersRegistry() {
        return typedi_1.default.get(ActionHandlersRegistry_1.ActionHandlersRegistry);
    }
    get templateUtilityRegistry() {
        return typedi_1.default.get(TemplateUtilitiesRegistry_1.TemplateUtilitiesRegistry);
    }
    /**
     * Extract idOrAlias from step object
     * @param {object} step
     * @return {string}
     */
    static extractIdOrAlias(step) {
        const keys = Object.keys(step);
        for (const key of keys) {
            if (!key.startsWith(FBLService_1.METADATA_PREFIX)) {
                return key;
            }
        }
        // this error should potentially never be thrown as all steps should be pre-validated
        throw new Error(`Unable to extract id or alias from keys: ${keys.join(', ')}`);
    }
    /**
     * Extract step metadata
     * @param {object} step
     * @return {IMetadata}
     */
    static extractMetadata(step) {
        const result = {};
        const keys = Object.keys(step);
        for (const key of keys) {
            if (key.startsWith(FBLService_1.METADATA_PREFIX)) {
                result[key] = step[key];
            }
        }
        return result;
    }
    /**
     * Get reporter by name
     * @param {string} name
     * @return {IReporter | undefined}
     */
    getReporter(name) {
        return this.reporters[name];
    }
    /**
     * Find plugin by its name
     * @param {string} name
     * @return {IPlugin | undefined}
     */
    getPlugin(name) {
        return this.plugins[name];
    }
    /**
     * Register single plugin
     * @param {IPlugin} plugin
     */
    registerPlugin(plugin) {
        this.plugins[plugin.name] = plugin;
        this.processedPlugins[plugin.name] = plugin.version;
        if (plugin.actionHandlers) {
            plugin.actionHandlers.forEach(actionHandler => {
                this.actionHandlersRegistry.register(actionHandler, plugin);
            });
        }
        if (plugin.reporters) {
            plugin.reporters.forEach(reporter => {
                this.reporters[reporter.getName()] = reporter;
            });
        }
        if (plugin.templateUtils) {
            this.templateUtilityRegistry.register(...plugin.templateUtils);
        }
    }
    /**
     * Register plugins
     * @param {IPlugin[]} plugins
     */
    registerPlugins(plugins) {
        plugins.forEach(plugin => {
            this.registerPlugin(plugin);
        });
    }
    /**
     * Try to require plugin, just return null if not found. Doesn't throw exceptions
     * @param {string} pluginNameOrPath
     * @param {boolean} name
     * @return {IPlugin | null}
     */
    static requirePluginSafe(pluginNameOrPath, name) {
        return __awaiter(this, void 0, void 0, function* () {
            let plugin = null;
            let exists = false;
            try {
                if (name) {
                    plugin = requireg(pluginNameOrPath);
                }
                else {
                    exists = yield utils_1.FSUtil.exists(pluginNameOrPath);
                    if (exists) {
                        plugin = require(pluginNameOrPath);
                    }
                    else {
                        exists = yield utils_1.FSUtil.exists(pluginNameOrPath + '.js');
                        if (exists) {
                            plugin = require(pluginNameOrPath);
                        }
                    }
                }
            }
            catch (e) {
                if (exists) {
                    throw e;
                }
                // ignore this
            }
            return plugin;
        });
    }
    /**
     * Require plugin
     * @param {string} pluginName
     * @param {string} wd
     * @return {IPlugin | null}
     */
    static requirePlugin(pluginName, wd) {
        return __awaiter(this, void 0, void 0, function* () {
            let plugin = yield FBLService_1.requirePluginSafe(pluginName, true);
            let childDir = null;
            if (!plugin) {
                while (!plugin && (!childDir || childDir !== wd)) {
                    // search in node_modules dir
                    plugin = yield FBLService_1.requirePluginSafe(path_1.resolve(wd, 'node_modules', pluginName), false);
                    /* istanbul ignore else */
                    if (!plugin) {
                        // search in directory with plugin name (useful when referencing sources)
                        plugin = yield FBLService_1.requirePluginSafe(path_1.resolve(wd, pluginName), false);
                    }
                    childDir = wd;
                    wd = path_1.resolve(wd, '..');
                }
            }
            if (!plugin) {
                throw new Error(`Unable to locate plugin ${pluginName}`);
            }
            return plugin;
        });
    }
    /**
     * Validate plugin to be registered and try to auto load if not
     * @param {string} pluginName
     * @param {string} pluginExpectedVersion
     * @param {string[]} errors
     * @param {string} wd working directory
     * @param {boolean} dryRun if true plugin just be verified, not actually registered
     */
    validateRequiredPlugin(pluginName, pluginExpectedVersion, errors, wd, dryRun) {
        return __awaiter(this, void 0, void 0, function* () {
            let version = this.processedPlugins[pluginName];
            if (!version) {
                try {
                    const plugin = yield FBLService_1.requirePlugin(pluginName, wd);
                    version = plugin.version;
                    this.processedPlugins[plugin.name] = plugin.version;
                    this.processedPlugins[pluginName] = plugin.version;
                    yield this.validatePlugin(plugin, wd);
                    /* istanbul ignore else */
                    if (!dryRun) {
                        this.registerPlugin(plugin);
                    }
                }
                catch (e) {
                    errors.push(`Required plugin ${pluginName.red} is not registered. Error: ${e.message}`);
                }
            }
            if (version && !semver.satisfies(version, pluginExpectedVersion)) {
                errors.push(`Actual plugin ${pluginName.red} version ${version.red} doesn't satisfy required ${pluginExpectedVersion.red}`);
            }
        });
    }
    /**
     * Validate plugin
     * @param {IPlugin} plugin
     * @param {string} wd
     */
    validatePlugin(plugin, wd) {
        return __awaiter(this, void 0, void 0, function* () {
            const errors = [];
            if (!semver.satisfies(fblVersion, plugin.requires.fbl)) {
                errors.push(`Plugin ${plugin.name.red} is not compatible with current fbl version (${fblVersion.red})`);
            }
            if (plugin.requires.plugins) {
                const dryRun = !this.allowUnsafePlugins && errors.length > 0;
                for (const dependencyPluginName of Object.keys(plugin.requires.plugins)) {
                    const dependencyPluginRequiredVersion = plugin.requires.plugins[dependencyPluginName];
                    yield this.validateRequiredPlugin(dependencyPluginName, dependencyPluginRequiredVersion, errors, wd, dryRun);
                }
            }
            if (plugin.requires.applications) {
                for (const application of plugin.requires.applications) {
                    let exists = false;
                    try {
                        exists = yield commandExists(application);
                    }
                    catch (e) {
                        exists = false;
                    }
                    if (!exists) {
                        errors.push(`Application ${application.red} required by plugin ${plugin.name.red} not found, make sure it is installed and its location presents in the PATH environment variable`);
                    }
                }
            }
            if (errors.length) {
                if (this.allowUnsafePlugins) {
                    errors.forEach(err => console.error(err));
                }
                else {
                    throw new Error(errors.join('\n'));
                }
            }
        });
    }
    /**
     * Validate plugins to be compatible
     * @param {string} wd
     */
    validatePlugins(wd) {
        return __awaiter(this, void 0, void 0, function* () {
            for (const name of Object.keys(this.plugins)) {
                yield this.validatePlugin(this.plugins[name], wd);
            }
        });
    }
    /**
     * Validate flow requirements
     * @param {IFlow} flow
     * @param {string} wd
     * @return {Promise<void>}
     */
    validateFlowRequirements(flow, wd) {
        return __awaiter(this, void 0, void 0, function* () {
            const errors = [];
            if (flow.requires) {
                if (flow.requires.fbl) {
                    if (!semver.satisfies(fblVersion, flow.requires.fbl)) {
                        errors.push(`Flow is not compatible with current fbl version (${fblVersion.red})`);
                    }
                }
                if (flow.requires.plugins) {
                    const dryRun = !this.allowUnsafeFlows && errors.length > 0;
                    for (const pluginName of Object.keys(flow.requires.plugins)) {
                        const pluginExpectedVersion = flow.requires.plugins[pluginName];
                        yield this.validateRequiredPlugin(pluginName, pluginExpectedVersion, errors, wd, dryRun);
                    }
                }
                if (flow.requires.applications) {
                    for (const application of flow.requires.applications) {
                        let exists = false;
                        try {
                            exists = yield commandExists(application);
                        }
                        catch (e) {
                            exists = false;
                        }
                        if (!exists) {
                            errors.push(`Application ${application.red} required by flow not found, make sure it is installed and its location presents in the PATH environment variable`);
                        }
                    }
                }
            }
            if (errors.length) {
                if (this.allowUnsafeFlows) {
                    errors.forEach(err => console.error(err));
                }
                else {
                    throw new Error(errors.join('\n'));
                }
            }
        });
    }
    /**
     * Execute flow
     * @param {string} wd Working Directory
     * @param {IFlow} flow,
     * @param {IContext} context
     * @param {IDelegatedParameters} parameters
     * @returns {Promise<ActionSnapshot>}
     */
    execute(source, wd, flow, context, parameters) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = Joi.validate(flow, schemas_1.FBL_FLOW_SCHEMA);
            if (result.error) {
                throw new Error(result.error.details.map(d => d.message).join('\n'));
            }
            yield this.validateFlowRequirements(flow, wd);
            return yield this.flowService.executeAction(source, wd, flow.pipeline, context, parameters);
        });
    }
};
FBLService.METADATA_PREFIX = '$';
FBLService = FBLService_1 = __decorate([
    typedi_1.Service()
], FBLService);
exports.FBLService = FBLService;
//# sourceMappingURL=FBLService.js.map