"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const glob = require("glob-promise");
const os_1 = require("os");
const path_1 = require("path");
const js_yaml_1 = require("js-yaml");
const util_1 = require("util");
const fs_1 = require("fs");
const errors_1 = require("../errors");
const unlinkAsync = util_1.promisify(fs_1.unlink);
const existsAsync = util_1.promisify(fs_1.exists);
const readdirAsync = util_1.promisify(fs_1.readdir);
const statAsync = util_1.promisify(fs_1.stat);
const rmdirAsync = util_1.promisify(fs_1.rmdir);
const readFileAsync = util_1.promisify(fs_1.readFile);
const renameAsync = util_1.promisify(fs_1.rename);
const copyFileAsync = util_1.promisify(fs_1.copyFile);
class FSUtil {
    /**
     * Check if path exists
     * @param {string} path
     * @return {Promise<boolean>}
     */
    static exists(path) {
        return __awaiter(this, void 0, void 0, function* () {
            return yield existsAsync(path);
        });
    }
    /**
     * Check if path is actually a URL
     * @param path
     */
    static isURL(path) {
        return path.indexOf('http://') === 0 || path.indexOf('https://') === 0;
    }
    /**
     * Check if path is absolute
     * @param {string} path
     * @return {boolean}
     */
    static isAbsolute(path) {
        return path_1.isAbsolute(path);
    }
    /**
     * Find files by mask
     * @param {string[]} masks
     * @param {string[]} ignore
     * @param {string} wd
     * @return {Promise<string[]>}
     */
    static findFilesByMasks(masks, ignore, wd) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = [];
            for (const mask of masks) {
                const absolutePathMask = FSUtil.getAbsolutePath(mask, wd);
                const matches = yield glob(absolutePathMask, {
                    ignore: (ignore || []).map(i => FSUtil.getAbsolutePath(i, wd)),
                    nodir: true,
                    absolute: true,
                    dot: true,
                });
                result.push(...matches);
            }
            return result;
        });
    }
    /**
     * Get absolute based on current working directory
     * @param {string} path
     * @param {string} wd Working Directory
     * @return {string}
     */
    static getAbsolutePath(path, wd) {
        // resolve path based on user home dir
        if (path.indexOf('~' + path_1.sep) === 0) {
            return path_1.resolve(os_1.homedir(), path.substring(2));
        }
        // if path is URL - return it back
        if (FSUtil.isURL(path)) {
            return path;
        }
        // if path is already absolute - return it back
        if (FSUtil.isAbsolute(path)) {
            return path_1.resolve(path);
        }
        // in any other case - resolve it as usual
        return path_1.resolve(wd, path);
    }
    static readFile(file) {
        return __awaiter(this, void 0, void 0, function* () {
            const fileExists = yield FSUtil.exists(file);
            if (!fileExists) {
                throw new errors_1.ActionError(`Unable to read file at path: ${file}. File is missing`, errors_1.NOT_FOUND);
            }
            return yield readFileAsync(file);
        });
    }
    /**
     * Get text file content
     * @param {string} file
     * @param {BufferEncoding} [encoding]
     * @return {Promise<string>}
     */
    static readTextFile(file, encoding = 'utf8') {
        return __awaiter(this, void 0, void 0, function* () {
            const data = yield FSUtil.readFile(file);
            return data.toString(encoding);
        });
    }
    /**
     * Create directory and all its parents recursively
     * @param {string} path
     * @return {Promise<void>}
     */
    static mkdirp(path) {
        return __awaiter(this, void 0, void 0, function* () {
            const exist = yield FSUtil.exists(path);
            if (!exist) {
                path = path_1.normalize(path);
                const parent = path_1.dirname(path);
                /* istanbul ignore else */
                if (parent !== path) {
                    yield this.mkdirp(parent);
                    yield util_1.promisify(fs_1.mkdir)(path);
                }
            }
            else {
                const directory = yield FSUtil.isDirectory(path);
                if (!directory) {
                    throw new errors_1.ActionError(`Unable to create folder at path: ${path}. Path already exists and it is not a folder.`, errors_1.PATH_ALREADY_EXISTS);
                }
            }
        });
    }
    /**
     * Remove files and folders at given path recursively
     * @param {string} path
     * @param {boolean} skipNotFound do not rise exception if file no longer exists
     * @return {Promise<void>}
     */
    static remove(path, skipNotFound = false) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!(yield existsAsync(path))) {
                if (skipNotFound) {
                    return;
                }
                throw new errors_1.ActionError(`Unable to find file or folder at path: ${path}`, errors_1.NOT_FOUND);
            }
            const directory = yield FSUtil.isDirectory(path);
            if (!directory) {
                yield unlinkAsync(path);
            }
            else {
                const children = yield readdirAsync(path);
                for (const child of children) {
                    yield FSUtil.remove(path_1.resolve(path, child));
                }
                yield rmdirAsync(path);
            }
        });
    }
    /**
     * Check if path represents a directory
     * @param {string} path
     * @return {Promise<boolean>}
     */
    static isDirectory(path) {
        return __awaiter(this, void 0, void 0, function* () {
            const stats = yield statAsync(path);
            return stats.isDirectory();
        });
    }
    /**
     * Move file/folder
     * @param {string} from
     * @param {string} to
     * @return {Promise<void>}
     */
    static move(from, to) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!(yield existsAsync(from))) {
                throw new errors_1.ActionError(`Unable to find file or folder at path: ${from}`, errors_1.NOT_FOUND);
            }
            const directory = yield FSUtil.isDirectory(from);
            if (!directory) {
                // if destination path has an OS specific path separator in the end
                if (to.endsWith(path_1.sep)) {
                    // create folders first
                    yield FSUtil.mkdirp(to);
                    // move file preserving its original name
                    yield renameAsync(from, path_1.resolve(to, path_1.basename(from)));
                }
                else {
                    // create folders first
                    yield FSUtil.mkdirp(path_1.dirname(to));
                    // move file ignoring its name
                    yield renameAsync(from, to);
                }
            }
            else {
                // if source path has an OS specific path separator in the end
                if (from.endsWith(path_1.sep)) {
                    // move its content instead of the folder itself
                    const children = yield readdirAsync(from);
                    for (const child of children) {
                        yield FSUtil.move(path_1.resolve(from, child), path_1.normalize(to) + path_1.sep);
                    }
                }
                else {
                    // if destination path has an OS specific path separator in the end
                    if (to.endsWith(path_1.sep)) {
                        // create folders first
                        yield FSUtil.mkdirp(to);
                        // move folder preserving its original name
                        yield renameAsync(from, path_1.resolve(to, path_1.basename(from)));
                    }
                    else {
                        // create folders first
                        yield FSUtil.mkdirp(path_1.dirname(to));
                        // move folder ignoring its name
                        yield renameAsync(from, to);
                    }
                }
            }
        });
    }
    /**
     * Copy file/folder
     * @param {string} from
     * @param {string} to
     * @return {Promise<void>}
     */
    static copy(from, to) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!(yield existsAsync(from))) {
                throw new errors_1.ActionError(`Unable to find file or folder at path: ${from}`, errors_1.NOT_FOUND);
            }
            const directory = yield FSUtil.isDirectory(from);
            if (!directory) {
                // if destination path has an OS specific path separator in the end
                if (to.endsWith(path_1.sep)) {
                    // create folders first
                    yield FSUtil.mkdirp(to);
                    // copy file preserving its original name
                    yield copyFileAsync(from, path_1.resolve(to, path_1.basename(from)));
                }
                else {
                    // create folders first
                    yield FSUtil.mkdirp(path_1.dirname(to));
                    // copy file ignoring its name
                    yield copyFileAsync(from, to);
                }
            }
            else {
                // if source path has an OS specific path separator in the end
                if (from.endsWith(path_1.sep)) {
                    // copy its content instead of the folder itself
                    const children = yield readdirAsync(from);
                    for (const child of children) {
                        yield FSUtil.copy(path_1.resolve(from, child), path_1.normalize(to) + path_1.sep);
                    }
                }
                else {
                    let destination;
                    // if destination path has an OS specific path separator in the end
                    if (to.endsWith(path_1.sep)) {
                        // create folders first
                        destination = path_1.resolve(to, path_1.basename(from));
                    }
                    else {
                        destination = to;
                    }
                    yield FSUtil.mkdirp(destination);
                    // copy its content instead of the folder itself
                    const children = yield readdirAsync(from);
                    for (const child of children) {
                        yield FSUtil.copy(path_1.resolve(from, child), destination + path_1.sep);
                    }
                }
            }
        });
    }
    /**
     * Read and parse yaml file
     * @param {string} file
     * @returns {Promise<any>}
     */
    static readYamlFromFile(file) {
        return __awaiter(this, void 0, void 0, function* () {
            const fileExists = yield FSUtil.exists(file);
            if (!fileExists) {
                throw new errors_1.ActionError(`Unable to find file at path: ${file}`, errors_1.NOT_FOUND);
            }
            const source = yield FSUtil.readTextFile(file);
            try {
                return js_yaml_1.safeLoad(source);
            }
            catch (e) {
                throw new errors_1.ActionError(`Unable to parse YAML: ${e.message}`, errors_1.CORRUPTED_FILE);
            }
        });
    }
}
exports.FSUtil = FSUtil;
//# sourceMappingURL=FSUtil.js.map